/***************************************************

Copyright 2003 Ben Childs


This file is part of Physics 3D.

    Physics 3D is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Physics 3D is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Physics 3D; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA




***************************************************/

package com.bchilds.Physics3D;


import javax.media.j3d.*;
import javax.vecmath.*;

import java.io.*;
import java.util.ArrayList;

import com.sun.j3d.utils.geometry.*;
import com.sun.j3d.utils.universe.*;
import com.sun.j3d.utils.image.TextureLoader;

//Explode behavior
import org.selman.java3d.book.behaviortest.*;

public class Particle3D implements ExplosionListener
{
	//Current time (in milliseconds) of the simulation
	private static long time = 0;
	
	//Collision Types
	public static final int EXPLODE = 1;
	public static final int MOMENTUM = 2;
	///////////////////////////////////////
	
	//What sort of collision should there be
	private static int collisionHandling = MOMENTUM;
	///////////////////////////////////////////////
	
	private static int explodeBlastRadiusMultiplier = 3;
	
	//Environmental Variables
	private static double gravity = 6.673E-11;
	///////////////////////////////////////////////////////
	
	//List of all particles in system
	private static ArrayList particles = new ArrayList();
	/////////////////////////////////////////////////////
	
	//Variables to keep track of tracking
	private static int traceNumFrames = -1;
	private static int currentFrame = -1;
	/////////////////////////////////////
	
	//Appearance for the tracks 
	private Appearance tracerAppearance;
	///////////////////////////////////
	
	//State variables for the particle
	private String name = null;
	
	private Vector3d velocityVector = null;
	private Vector3d initialVelocityVector = null;
	
	private Vector3d angularVelocity = new Vector3d(0,0,0);
	private Vector3d initialAngularVelocity = new Vector3d(0,0,0);
	
	private Vector3d position =  null;
	private Vector3d initialPosition = null;
	private Vector3d lastPosition = null;
	
	private Quat4d orientation = makeQFromEulerAngles(0,0,0);
	private Quat4d initialOrientation = makeQFromEulerAngles(0,0,0);
	private Quat4d lastOrientation = null;
	
	private Matrix3d momentOfInertia = null;
	
	private double mass = 1;
	private double radius = .1;
	private double elasticity = 1; //Between 0 and 1	
	///////////////////////////////////////
	
	//For updates if true use last position, otherwise use current position
	//in calculation of acceleration
	private boolean updated = false;
	/////////////////////////////////////
	
	
	
	
	//Java3D objects representing the particle
	private TransformGroup object3D = null; //The transform group for setting position/size
	private Shape3D shape3D = null;			//The shape for exploding
	private BranchGroup bg = null;			//The branch group for removing from the 
	
	private BranchGroup bgTracer = null;		//Branch group for holding all the tracer objects
	private static BranchGroup bgRoot = null;	//The root branch group, what we remove stuff from...
	private Point3d[] geom = null;				//All the vertices for the object
	private ArrayList p3dadj = null;			//The adjacent table
	private Feature[] closestFeatures = null;
	
	
	//Bounding Box Stuff
	private static Point3d ufrBounding = null;
	private static Point3d lblBounding = null;
	private static double boundsElasticity = 1;
	/////////////////
	
	
			
	//Creates a Particle3D object given a Java3D Transform group, a velocity vector and the mass
	public Particle3D(String name, TransformGroup object, Vector3d velocity, double mass, Vector3d initialPosition, BranchGroup bGroup, Shape3D shape, double elasticity)
	{
		this.name = name;
		setSpeed(velocity);
		this.mass = mass;
		object3D = object;
		this.elasticity = elasticity;
		setPosition(initialPosition);
		bgTracer = new BranchGroup();
		bgTracer.setCapability(BranchGroup.ALLOW_CHILDREN_EXTEND);
		bgTracer.setCapability(BranchGroup.ALLOW_DETACH);		
		bg = bGroup;
		bg.addChild(bgTracer);
		shape3D = shape;
		tracerAppearance = (Appearance)shape3D.getAppearance().cloneNodeComponent();
		tracerAppearance.setTransparencyAttributes(new TransparencyAttributes(TransparencyAttributes.NICEST, .75f));
		particles.add(this);
		
		
		//Calculate Moment of Inertia (standard formula for sphere
		double ixx, iyy, izz;
		
		ixx = (2/5.0) * mass * radius * radius;
		iyy = ixx;
		izz = ixx;
		momentOfInertia = new Matrix3d(ixx, 0, 0, 0, iyy, 0, 0, 0, izz);
		/////////////////////////////
		
		//Calculate adjacent vertices (for new Collision detection scheme)
		
		//Get the vertices
		GeometryArray g = (GeometryArray)shape3D.getGeometry();
		g.setCapability( GeometryArray.ALLOW_COORDINATE_READ );
		g.setCapability( GeometryArray.ALLOW_COORDINATE_WRITE );
		g.setCapability( GeometryArray.ALLOW_COUNT_READ );
		
		geom = new Point3d[g.getVertexCount()];
		for(int i = 0; i < geom.length; i++)
		{
			geom[i] = new Point3d(0.0,0.0,0.0);
		}
		g.getCoordinates(0, geom);
		///////////////////////////
		
		//Set large negative doubles to 0
		for(int i = 0; i < geom.length; i++)
		{
			if(Math.abs(geom[i].x) < 1E-10)
				geom[i].x = 0.0;
			if(Math.abs(geom[i].y) < 1E-10)
				geom[i].y = 0.0;
			if(Math.abs(geom[i].z) < 1E-10)
				geom[i].z = 0.0;		
		}
		/////////////////////////////////////
		
		p3dadj = new ArrayList();	//Initialize the arraylist of vertices/adjacents
		
		ArrayList p3dKeys = new ArrayList(); //Temporary Arraylist to keep the adjacent particles 
		
		//Iterate through all the vertices
		for(int i = 0; i < geom.length; i++)
		{
			//Check to see if the vertex has already been done
			boolean unique = true;
			for(int m = 0; m < p3dadj.size(); m++)
			{
				if(((Point3DAdjacent)p3dadj.get(m)).equals(geom[i]))
				{
					unique = false;
					break;
				}
			}
			if(!unique)
				continue;
			//////////////////////////////////////////////////////
			
			//List of equal vertices	
			ArrayList l = new ArrayList();
			
			for(int j = i; j < geom.length; j++)
			{
				//Get all equal.
				
				if(geom[i].x == geom[j].x && 
					geom[i].y == geom[j].y &&
					geom[i].z == geom[j].z)
					l.add(new Integer(j));
			}
			///////////////////////////////
			
			//System.out.println("Found " + l.size() + " identical points at positions:");
			//for(int j = 0; j < l.size(); j++)
			//{
			//	System.out.print(l.get(j) + " : ");
			//}
			//System.out.println();
			
			//List of indices of adjacent particles
			ArrayList adjacent = new ArrayList();
			
			//Cycle through equal vertices
			for(int k = 0; k < l.size(); k++)
			{
				//Get the current particle to test
				int current = ((Integer)l.get(k)).intValue();
				
				//Either the particle is the middle of a triangle, or the top/bottom, copy appropriately
				if(current > 0 && current < geom.length -1 &&
					 current % 3 == 1)
				{
				//	System.out.println("Going up and down from pt:" + current);
					adjacent.add(new Integer(current-1));
					adjacent.add(new Integer(current+1));
				}	
				else if(current > 1 && current % 3 == 2)
				{
				//	System.out.println("Going up from pt:" + current);
					adjacent.add(new Integer(current-1));
					adjacent.add(new Integer(current-2));
				}
				else if(current < geom.length - 2 && current % 3 == 0)
				{
				//	System.out.println("Going down from pt:" + current);
					adjacent.add(new Integer(current+1));
					adjacent.add(new Integer(current+2));
				}
				
				//remove duplicates
			//	removeDupP3d(adjacent, geom);
			}
			
			//Adjacent keys...
			Point3DAdjacent pa = new Point3DAdjacent();
			pa.point = geom[i];
			p3dadj.add(pa);
			Point3d p3dadjacents[] = new Point3d[adjacent.size()];
			
			//System.out.println("For Point: " + geom[i] + " calculated adjacents to be:");
			for(int n = 0; n < p3dadjacents.length; n++)
			{
				p3dadjacents[n] = geom[((Integer)adjacent.get(n)).intValue()];
			//	System.out.println(p3dadjacents[n]);
			}

			p3dKeys.add(p3dadjacents);
		}
		
		//Calculate the keys and add them to the adjacent table
		for(int i = 0; i < p3dadj.size() && i < p3dKeys.size(); i++)
		{
			Point3DAdjacent pa = (Point3DAdjacent)p3dadj.get(i);
			
			Point3d[] adjacents = (Point3d[])p3dKeys.get(i);
			
			int[] iAdj = new int[adjacents.length];
			
			//System.out.println("Calculated Keys for : " + pa.point );
			for(int j = 0; j < adjacents.length; j++)
			{
				
				for(int k = 0; k < p3dadj.size(); k++)
				{
					if(p3dadj.get(k).equals(adjacents[j]))
					{
						iAdj[j] = k;
						break;
					}
				}
				//System.out.print(iAdj[j] + ",");
			}
			//System.out.println();
			pa.key = iAdj;

			
		}
		
		
	/*	
		BranchGroup bgCollision = new BranchGroup();
		CollisionBehavior cb = new CollisionBehavior(this);
		bgCollision.addChild(cb);
		bg.addChild(bgCollision);
	*/	
		
	}
	
	//Removes duplicate keys from the arraylist (used for calculation of adjacent vertices)
	private void removeDupP3d(ArrayList keys, Point3d[] p3ds)
	{
		for(int i = 0; i < keys.size(); i++)
		{
			Point3d current = p3ds[((Integer)keys.get(i)).intValue()];
			
			for(int j = i+1; j < keys.size(); j++)
			{
				Point3d test = p3ds[((Integer)keys.get(j)).intValue()];
				if(current.x == test.x &&
					current.y == test.y &&
					current.z == test.z)
				{
					keys.remove(j);
					j--;
				}
					
			}
		}
	}


	private Quat4d quatMult43(Quat4d v4, Vector3d v3)
	{
		return new Quat4d(		v4.w * v3.x + v4.y * v3.z - v4.z * v3.y,
								v4.w * v3.y + v4.z * v3.x - v4.x * v3.z,
								v4.w * v3.z + v4.x * v3.y - v4.y * v3.x,
								-(v4.x * v3.x + v4.y * v3.y + v4.z * v3.z));
		
	
	}
	
	//Update the objects position based on some timeDifferential (time change to represent)
	private boolean update(float timeDifferential)
	{
				
		//Calculate the acceleration on the object
		Vector3d accelerationVector = calculateAcceleration(this);
				
								
		//Calculate the change in velocity given the time differential
		Vector3d velocityChange = accelerationVector;
		velocityChange.scale(timeDifferential);
		
		//Update the velocity vector
		velocityVector.add(velocityChange);
		
		//Calculate the change in position
		Vector3d positionChange = (Vector3d)velocityVector.clone();
		positionChange.scale(timeDifferential);
		
		//Store last position
		lastPosition = position;
		
		//Update current position
		position.add(positionChange);
	
		
		
		//Calculate new Orientation from Angular Velocity
		Quat4d newOrientation;
	
		newOrientation = quatMult43(orientation, angularVelocity);
		newOrientation.scale(.5 * timeDifferential);
		
		newOrientation.add(orientation);
		
		lastOrientation = orientation;
		orientation = newOrientation;
		orientation.normalize();
		
		
		//Update the transform group
		Transform3D updateTransform = new Transform3D();
		updateTransform.setTranslation(position);
		updateTransform.setScale(radius*10);
		updateTransform.setRotation(newOrientation);
		object3D.setTransform(updateTransform);
		
		//If we are tracing and it is the right frame add trace object
		if(currentFrame == 0)
		{
			TransformGroup tgTracer = new TransformGroup();
			Transform3D t3d = new Transform3D();
			t3d.setScale(radius*10);
			t3d.setTranslation(position);
			t3d.setRotation(newOrientation);
			tgTracer.setTransform(t3d);
			tgTracer.addChild(new Sphere(0.1f, Primitive.GENERATE_NORMALS, 20, tracerAppearance));
			BranchGroup bgNew = new BranchGroup();
			bgNew.addChild(tgTracer);
			bgTracer.addChild(bgNew);
		}
		
		updated = true;
		
		return true;
		
	}
	
	
	
	//Calculate the Acceleration due to gravity of some Particle3D
	private static Vector3d calculateAcceleration(Particle3D p)
	{
		Vector3d totalAcceleration = new Vector3d(0.0,0.0,0.0); //Total acceleration so far
		
		if(gravity == 0)			//Efficiency test (avoids an O(n) operation if needless)
			return totalAcceleration;
		
		//Iterate through all particles and add the acceleration of all of them	
		int length = particles.size();		
		for(int i = 0; i < length; i++)
		{
			totalAcceleration.add(calculateAttraction2Particles(p, (Particle3D)particles.get(i)));
		}
		/////////////////////////////////////
		
		return totalAcceleration;
	
	}

	//Calculate the acceleration on p1 due to p2.
	private static Vector3d calculateAttraction2Particles(Particle3D p1, Particle3D p2)
	{
		if(p1.equals(p2))
		{
		//	System.out.println("Got same particles, aborting");
			return new Vector3d(0.0,0.0,0.0);
		}
		//Get the current positions of the particles (for this update cycle)		
		Vector3d  	positionp1 = (p1.updated ? p1.lastPosition : p1.position),
					positionp2 = (p2.updated ? p2.lastPosition : p2.position);
		
		//Get distance between them					
		Vector3d vdistance = (Vector3d)positionp2.clone();
		vdistance.sub(positionp1);
		double fdistance = vdistance.length();
		
		//Get mass of p2
		double mass = p2.mass;
		
		//Formula for acceleration due to gravity
		double acceleration = (gravity * mass) / (fdistance * fdistance);
		////////////////////////////////////////
		
		//Get the acceleration vector
		Vector3d vAcceleration = vdistance;
		vAcceleration.scale((double)(acceleration / fdistance));
		/////////////////////////////////////
		
		return vAcceleration;
	}
	
	
	////////EXPLOSION STUFF/////////////////	
	BranchGroup explodeBg;
	public void explode()
	{
		ExplodeBehavior exp = new ExplodeBehavior(shape3D, 10, 60, this, 50);
		exp.setSchedulingBounds(Physics3DThread.getBoundingSphere());
		explodeBg = new BranchGroup();
		explodeBg.setCapability(BranchGroup.ALLOW_DETACH);
		explodeBg.addChild(exp);
		bg.addChild(explodeBg);

	}
	
	public WakeupCondition onExplosionFinished( ExplodeBehavior explodeBehavior, Shape3D shape3D )
	{
		bgRoot.removeChild(bg);
		return null;
	}
	///////////////////////////////////////
	
	
	///////Collision Stuff/////////

	private static Vector3d multMatrixVector(Matrix3d m, Vector3d v)
	{
		return new Vector3d( m.m00 * v.x + m.m01 * v.y + m.m02 * v.z,
								m.m10 * v.x + m.m11 * v.y + m.m12 * v.z,
								m.m20 * v.x + m.m21 * v.y + m.m22 * v.z);
	
	}
	
	//Calculate the resultantVelocity of a collision between particles 1 and 2
	public static void calculateResultantVelocities(Particle3D p1, Particle3D p2, int time, int i1, int i2)
	{
		//Get current velocities
		Vector3d p1vector = (Vector3d)p1.velocityVector.clone();
		Vector3d p2vector = (Vector3d)p2.velocityVector.clone();
		Vector3d p1avector = (Vector3d)p1.angularVelocity.clone();
		Vector3d p2avector = (Vector3d)p2.angularVelocity.clone();
		///////////////////////
		
		Vector3d relPosition1 = new Vector3d(p1.closestFeatures[i2].getPoint());
		relPosition1.sub(p1.position);
		
				
		Vector3d relPosition2 = new Vector3d(p2.closestFeatures[i1].getPoint());
		relPosition2.sub(p2.position);
		
		
		//Calculate relative velocity
		Vector3d relativeVelocity = (Vector3d)p1vector.clone();
		
		Vector3d angRel = (Vector3d)p1.angularVelocity.clone();
		angRel.cross(angRel, relPosition1);
		
		Vector3d angRel2 = (Vector3d)p2.angularVelocity.clone();
		angRel2.cross(angRel2, relPosition2);
		
		relativeVelocity.add(angRel);
		
		relativeVelocity.sub(p2vector);
		
		relativeVelocity.sub(angRel2);
		
		
		////////////////////////
		
		//Calculate normal of collision (needs to be updated if new shapes are introduced)
		Vector3d normal = (Vector3d)p1.position.clone();
		normal.sub(p2.position);
		
		normal.normalize();
		///////////////////////////////////////////////////////////////////
		
		//Calculate elasticity of the collision (is this correct?)
		double e = (p1.elasticity + p2.elasticity) / 2;
		//////////////////////////////////////////////////////
		
		//Caculate the impulse (Thank you to Chris Hecker for his great article about
		//Physics in Games avaliable at: http://www.d6.com/users/checker
		Vector3d temp = (Vector3d)relativeVelocity.clone();
		temp.scale((-(1 + e)));
		
		Matrix3d p1moiInverse = (Matrix3d)p1.momentOfInertia.clone();
		p1moiInverse.invert();
		
		Matrix3d p2moiInverse = (Matrix3d)p2.momentOfInertia.clone();
		p2moiInverse.invert();
		
		double tempJ = ((1 / p1.mass) + (1 / p2.mass));
		
		Vector3d temp2 = (Vector3d)relPosition1.clone();
		temp2.cross(temp2, normal);
		temp2 = multMatrixVector(p1moiInverse, temp2);
		temp2.cross(temp2, relPosition1);
		tempJ += temp2.dot(normal);
		
		temp2 = (Vector3d)relPosition2.clone();
		temp2.cross(temp2, normal);
		temp2 = multMatrixVector(p2moiInverse, temp2);
		temp2.cross(temp2, relPosition2);
		tempJ += temp2.dot(normal);
		
		double j = temp.dot(normal) / tempJ;
		
		//Reverse the elasticity if rewinding
		if(time == -1)
			j /= -e;
		///////////////////////////////////////////////
		
		System.out.println("Collision between " + p1.name + " and " + p2.name + " impulse = " + j);
		
		
		//Calculate the resultant velocities
		Vector3d temp3 = (Vector3d)normal.clone();
		temp3.scale((j / p1.mass));
		
		if(time == 1)
			p1vector.add(temp3);
		else
			p1vector.sub(temp3);
		
		
		temp3 = (Vector3d)normal.clone();
		temp3.scale(j / p2.mass);

		if(time == 1)
			p2vector.sub(temp3);
		else
			p2vector.add(temp3);
			
		
		temp3 = (Vector3d)normal.clone();
		temp3.scale(j);
		temp3.cross(temp3, relPosition1);
		p1.momentOfInertia.transform(temp3);
		
		if(time == 1)
			p1avector.add(temp3);
		else
			p1avector.sub(temp3);
		
		temp3 = (Vector3d)normal.clone();
		temp3.scale(j);
		temp3.cross(temp3, relPosition2);
		p2.momentOfInertia.transform(temp3);
		
		if(time == 1)
			p2avector.add(temp3);
		else
			p2avector.sub(temp3);
		
		p1.velocityVector = p1vector;
		p1.angularVelocity = p1avector;
		
		p2.velocityVector = p2vector;
		p2.angularVelocity = p2avector;
		//////////////////////////////
	
	
	}	
	
	
	//////////////////////////////
	
	//Are the two particles colliding? (used by updateAll function);
	public static boolean areColliding(Particle3D p1, Particle3D p2, int time, int i1, int i2)
	{
		
		//Quick special case check
		if(p1.equals(p2))
		{
			return false;
		}
		
	//	if(p1.closestFeatures[i2].distanceTo(p2.position) < p2.closestFeatures[i1].distanceTo(p2.position))
	//	{
	//		System.out.println("CollisionHandling 2 Detected Collision: " + p1 + " : " + p2);
	//	}
		
		
		//Get the current positions of the particles	
	//	Vector3d	positionp1 =  p1.position,
	//				positionp2 = p2.position;
							
	//	Vector3d vdistance = (Vector3d)positionp2.clone();
	//	vdistance.sub(positionp1);
			
	//	double fdistance = vdistance.length();
		
		//If they are close enough check the relative velocities
	//	if(Math.abs(fdistance) <= Math.abs(p1.radius + p2.radius))
		if(p1.closestFeatures[i2].distanceTo(p2.position) <= p2.closestFeatures[i1].distanceTo(p2.position))
		{
	
			Vector3d	velocity1 = (Vector3d)p1.velocityVector.clone(),
						velocity2 = (Vector3d)p2.velocityVector.clone();
							
			Vector3d relativeVelocity = velocity1;
			relativeVelocity.sub(velocity2);
		
		
			Vector3d normal = (Vector3d)p1.position.clone();
			normal.sub(p2.position);
			
			//Vector3d normal = Feature.calculateNormal(p1.closestFeatures[i2], p2.closestFeatures[i1]);
			
			
			//Return if they are colliding
			return  (time == 1 ? relativeVelocity.dot(normal) < 0 : relativeVelocity.dot(normal) > 0);
			
		}
		return false;
	
	}
	
		
	public static boolean collideWithBounds(Particle3D p)
	{
		//If there are no bounds skip out
		if(ufrBounding == null || lblBounding == null)
			return false;
		
		//Get the normal Vector of collision
		Vector3d normal = new Vector3d(0.0,0.0,0.0);
		
		normal.x = (ufrBounding.x - p.position.x < p.radius ? 1.0 : 
								(p.position.x - lblBounding.x < p.radius ? -1.0 : 0));
		normal.y = (ufrBounding.y - p.position.y < p.radius ? 1.0 : 
								(p.position.y - lblBounding.y < p.radius ? -1.0 : 0));
		normal.z = (ufrBounding.z - p.position.z < p.radius ? 1.0 : 
								(p.position.z - lblBounding.z < p.radius ? -1.0 : 0));						

		//If there isn't a collision skip out
		if(normal.length() == 0)
			return false;
		
		//Get the relative velocity
		Vector3d pvector = (Vector3d)p.velocityVector.clone();
		Vector3d relativeVelocity = (Vector3d)pvector.clone();		
		
		//Get elasticiyty
		double e = (p.elasticity + boundsElasticity) / 2;
		
		//Get impulse
		Vector3d temp = (Vector3d)relativeVelocity.clone();
		temp.scale((-1 * (1 + e)));
		
		Vector3d temp2 = (Vector3d)normal.clone();
		temp2.scale((1 / p.mass));
		
		double j = temp.dot(normal) / normal.dot(temp2);
		//////////////////////
		
		
		System.out.println("Collision between " + p.name + " and bounding wall: impulse = " + j);
		
		//Calculate resultant velocity
		Vector3d temp3 = (Vector3d)normal.clone();
		temp3.scale((j / p.mass));
		
		pvector.add(temp3);
	
		p.velocityVector = pvector;
		
		return true;
			
	}
	
	//Handles the collision of 2 particles
	public static boolean handleCollision(Particle3D p1, Particle3D p2, int time, int i1, int i2)
	{
		
		//if they aren't colliding don't do anything
		if(!areColliding(p1, p2, time, i1, i2))
			return false;
		
		System.out.println("Collision: p1: " + p1.name + " p2: " + p2.name);
		
		//Handle exploding...
		if(collisionHandling == EXPLODE)
		{
			p1.explode();
			p2.explode();
			
			particles.remove(p1);
			particles.remove(p2);
			
			for(int i = 0; i < particles.size(); i++)
			{
				Particle3D current = (Particle3D)particles.get(i);
				
				Vector3d distance = new Vector3d(0.0,0.0,0.0);
				distance.add(current.position);
				distance.sub(p1.position);
				
				if(Math.abs(distance.length()) < p1.radius * explodeBlastRadiusMultiplier)
				{
					current.explode();
					particles.remove(current);
					i = 0;
				}
			}
		
		}
		//Handle momentum
		else if(collisionHandling == MOMENTUM)
			calculateResultantVelocities(p1, p2, time, i1, i2);

		return true;
	}
	
	//Removes a particle from the system
	public static void removeParticle(Particle3D p)
	{
		Particle3DTracker.removeReferring(p);
		particles.remove(p);
		bgRoot.removeChild(p.bg);
	
	}
	
	public static void setup()
	{
		int length = particles.size();
		for(int i = 0; i < length; i++)
		{
			((Particle3D)particles.get(i)).closestFeatures = new Feature[length];
		
		}
		
		//Get Closest Features
		for(int i = 0; i < length; i++)
		{
			for(int j = i+1; j < length; j++)
			{
				getClosest(i, j);
			}
		}
		/////
	
	}
	
	//Updates all particles in the system
	public static void updateAll(float timeDifferential)
	{
		//Get the time
		time += (long)(timeDifferential * 1000);
		int normalAdjust = 1;

		if(timeDifferential < 0)
			normalAdjust *= -1;
		//////////////////////
		
		//Update all the particles positions and collide with bounds (O(n))
		int length = particles.size();
				
		for(int i = 0; i < length; i++)
		{
			((Particle3D)particles.get(i)).update(timeDifferential);
			collideWithBounds(((Particle3D)particles.get(i)));
		}
		
		
		//Get Closest Features
		for(int i = 0; i < length; i++)
		{
			for(int j = i+1; j < length; j++)
			{
				getClosest(i, j);
			}
		}
		/////
		
		//Collide particles with each other (O(n^2) :()
		for(int i = 0; i < length; i++)
		{
			
			Particle3D current = (Particle3D)particles.get(i);
			for(int j = 0; j < length; j++)
			{					
				if(handleCollision(current, (Particle3D)particles.get(j), normalAdjust, i, j))
				{
					if(collisionHandling == EXPLODE)
					{
						length = particles.size();
						i--;
						break;				
					}
				
				}
			}
		}
		
		//Set updated back to false
		for(int i = 0; i < length; i++)
		{
			((Particle3D)particles.get(i)).updated = false;
		}
		
		//If we are traking update the current frame
		if(traceNumFrames != -1)
		{
			currentFrame = ((currentFrame >= traceNumFrames) ? 0 : currentFrame + 1);
		}
		
		//Update all the trackers
		Particle3DTracker.updateAll();
				
	}
	
	public static void getClosest(int i1, int i2)
	{
		Particle3D 	p1 = (Particle3D)particles.get(i1),
					p2 = (Particle3D)particles.get(i2);
		
		if(p1.closestFeatures[i2] == null)
		{
			p1.closestFeatures[i2] = p1.new Feature(((Point3DAdjacent)p1.p3dadj.get(0)).point, 0);
		}
		if(p2.closestFeatures[i1] == null)
		{
			p2.closestFeatures[i1] = p2.new Feature(((Point3DAdjacent)p2.p3dadj.get(0)).point, 0);
		}
		
		while(p1.getClosestFeatureTo(p2.closestFeatures[i1], i2) ||
				p2.getClosestFeatureTo(p1.closestFeatures[i2], i1));

	}
	
	public boolean getClosestFeatureTo(Feature f, int i)
	{
		Feature current = closestFeatures[i];
		
		double distance = current.distanceTo(f);
		
		Feature[] adj = null;
		
		boolean changed = true;
		boolean retVal = false;
		while(changed)
		{
			changed = false;
			adj = current.getAdjacent();
			
			for(int j = 0; j < adj.length; j++)
			{	
				double d = adj[j].distanceTo(f);
				if(d < distance)
				{
					distance = d;
					current = adj[j];
					changed = true;
					retVal = true;
				}
			}
		}
		
		closestFeatures[i] = current;
		return retVal;
	}
	
	public static void resetAll()
	{
		time = 0;
		int length = particles.size();
				
		for(int i = 0; i < length; i++)
		{
			((Particle3D)particles.get(i)).resetPosSpeed();
		}
		Particle3DTracker.updateAll();
		
	}
	
	
	
	//Resets posiiton and speed of particles to their initial position
	public void resetPosSpeed()
	{
		System.out.println("Resetting " + name + " to: " + initialPosition);
		setPosition(initialPosition);
		setSpeed(initialVelocityVector);
		setOrientation(initialOrientation);
		setAngularSpeed(initialAngularVelocity);
		bg.removeChild(bgTracer);
		
		bgTracer = new BranchGroup();
		bgTracer.setCapability(BranchGroup.ALLOW_CHILDREN_EXTEND);
		bgTracer.setCapability(BranchGroup.ALLOW_DETACH);		
		bg.addChild(bgTracer);
		
	}
	
	//Removes all particles from the system	
	public static void removeAll()
	{
		while(particles.size() > 0)
		{
			removeParticle((Particle3D)particles.get(0));
		}
	}
	

	/* File format:
	
	Gravity
	TraceNumFrames
	Number of Particles
	UFRBounding point x
	y
	z
	LBLBounding point x
	y
	z
	for each particle: name
	position x
	y
	z
	velocity x
	y
	z
	mass
	radius
	elasticity
	
	*/

	//Functions for loading and saving the particle to file...
	public static void saveAll(BufferedWriter w) throws IOException
	{
		w.write(gravity + "\n");
		w.write(traceNumFrames + "\n");
		w.write(particles.size() + "\n");
		if(ufrBounding == null || lblBounding == null)
		{
			w.write("null\nnull\nnull\nnull\nnull\nnull\n");
		}
		else
		{
			w.write(ufrBounding.x + "\n");
			w.write(ufrBounding.y + "\n");
			w.write(ufrBounding.z + "\n");
			w.write(lblBounding.x + "\n");
			w.write(lblBounding.y + "\n");
			w.write(lblBounding.z + "\n");
		}
		
		for(int i = 0; i < particles.size(); i++)
		{
			Particle3D current = (Particle3D)particles.get(i);
			w.write(current.name + "\n");
			w.write(current.initialPosition.x + "\n");
			w.write(current.initialPosition.y + "\n");
			w.write(current.initialPosition.z + "\n");
			w.write(current.initialVelocityVector.x + "\n");
			w.write(current.initialVelocityVector.y + "\n");
			w.write(current.initialVelocityVector.z + "\n");
			w.write(current.mass + "\n");
			w.write(current.radius + "\n");
			w.write(current.elasticity + "\n");
		}
	}
	
	//Load the particles from a buffered reader
	public static void loadAll(BufferedReader r, Physics3DThread t) throws IOException
	{
		setGravity(Double.parseDouble(r.readLine()));
		setTraceNumFrames(Integer.parseInt(r.readLine()));
		int length = Integer.parseInt(r.readLine());
		
		String line = r.readLine();
		//Read the bounding box
		if(line.equals("null"))
		{
			r.readLine();
			r.readLine();
			r.readLine();
			r.readLine();
			r.readLine();
			ufrBounding = null;
			lblBounding = null;
		}
		else
		{
			ufrBounding = new Point3d(Double.parseDouble(line),
										Double.parseDouble(r.readLine()),
										Double.parseDouble(r.readLine()));
			lblBounding = new Point3d(Double.parseDouble(r.readLine()),
										Double.parseDouble(r.readLine()),
										Double.parseDouble(r.readLine()));
		}
		////////////////////////
		
		//Read all the particles in
		for(int i = 0; i < length; i++)
		{
			String name = r.readLine();
			Vector3d position = new Vector3d(Double.parseDouble(r.readLine()),
											Double.parseDouble(r.readLine()),
											Double.parseDouble(r.readLine()));
			Vector3d velocity = new Vector3d(Double.parseDouble(r.readLine()),
											Double.parseDouble(r.readLine()),
											Double.parseDouble(r.readLine()));
			double mass = Double.parseDouble(r.readLine());
			double radius = Double.parseDouble(r.readLine());
			double elasticity = Double.parseDouble(r.readLine());
			
			t.addParticle(name, position, velocity, mass, radius, elasticity);
		
		}
		////////////////////////////////
	}
	
	//Accessor / Setter functions
	public static void setGravity(double g)
	{
		gravity = g;
	}
	
	public static double getGravity()
	{
		return gravity;
	}
	
	public void setRadius(double newRadius)
	{
		
		Transform3D newTransform = new Transform3D();
		object3D.getTransform(newTransform);
		radius = newRadius;
		newTransform.setScale(radius * 10);
		object3D.setTransform(newTransform);
		
		
	}
	
	public double getRadius()
	{
		return radius;
	}
	
	
	public double getMass()
	{
		return mass;
	}
	
	public Vector3d getPosition()
	{
		return position;
	}
	
	public Point3d getPositionP3d()
	{
		return new Point3d(position.x, position.y, position.z);
	}
	
	public Vector3d getSpeed()
	{
		return velocityVector;
	}
	
	public void setPosition(Vector3d newPosition)
	{
		Transform3D newPositionT3d = new Transform3D();
		object3D.getTransform(newPositionT3d);
		newPositionT3d.setTranslation(newPosition);
		
		object3D.setTransform(newPositionT3d);
		
		lastPosition = null;

		initialPosition = (Vector3d)newPosition.clone();
		position = newPosition;
	}
	
	public Quat4d getOrientation()
	{
		return (Quat4d)orientation.clone();
	}
	
	public void setOrientation(Quat4d newOrientation)
	{
		Transform3D newOrientationT3d = new Transform3D();
		object3D.getTransform(newOrientationT3d);
		newOrientationT3d.setRotation(newOrientation);
		
		object3D.setTransform(newOrientationT3d);
		
		lastOrientation = null;

		initialOrientation = (Quat4d)newOrientation.clone();
		orientation = newOrientation;
	
	}
	
	public void setSpeed(Vector3d newSpeed)
	{
		velocityVector = newSpeed;
		initialVelocityVector = (Vector3d)newSpeed.clone();
	}
	
	public Vector3d getAngularSpeed()
	{
		return (Vector3d)angularVelocity.clone();
	}
	
	public void setAngularSpeed(Vector3d newASpeed)
	{
		angularVelocity = newASpeed;
		initialAngularVelocity = (Vector3d)newASpeed.clone();
	}
	
	public void setMass(double newMass)
	{
		mass = newMass;
	}
	
	public String getName()
	{
		return name;
	}
	
	public void setName(String newName)
	{
		name = newName;
	}
	
	public double getElasticity()
	{
		return elasticity;
	}
	
	public void setElasticity(double newElasticity)
	{
		elasticity = newElasticity;
	}
	
	public static void setTraceNumFrames(int newTraceNumFrames)
	{
		traceNumFrames = newTraceNumFrames;
		if(traceNumFrames != -1)
			currentFrame = 0;
		else
			currentFrame = -1;
	}
	
	public static int getTraceNumFrames()
	{
		return traceNumFrames;
	}
	
	public static void setRoot(BranchGroup theBgRoot)
	{
		bgRoot = theBgRoot; 
	}
	
	public static Object[] getParticles()
	{
		return particles.toArray();
	}
	
	public String toString()
	{
		return name;
	}
	
	public static Particle3D getNextParticle(Particle3D p)
	{
		int index = particles.indexOf(p);
		index++;
		
		if(index == particles.size())
			index = 0;
			
		return (Particle3D)particles.get(index);
	}
	
	public static Particle3D getPrevParticle(Particle3D p)
	{
		int index = particles.indexOf(p);
		index--;
		
		if(index == -1)
			index = particles.size() - 1;
			
		return (Particle3D)particles.get(index);
	}
	
	public static void setBoundingBox(Point3d ufr, Point3d lbl)
	{
		ufrBounding = ufr;
		lblBounding = lbl;
	}
	
	public static Point3d getUFRBounding()
	{
		return ufrBounding;
	}
	
	public static Point3d getLBLBounding()
	{
		return lblBounding;
	}
		
	public void setAppearance(Appearance app)
	{
		shape3D.setAppearance(app);
	}
	
	public Shape3D getShape3D()
	{
		return shape3D;
	}
	
	public Vector3d getMomentum()
	{
		Vector3d v = ((Vector3d)velocityVector.clone());
		v.scale(mass);
		return v;
	}
	
	public double getKineticEnergy()
	{
		return velocityVector.lengthSquared() * mass;
	}
	
	public static void setCollisionHandling(int newCollisionHandling)
	{
		if(newCollisionHandling == MOMENTUM || newCollisionHandling == EXPLODE)
			collisionHandling = newCollisionHandling;
	}
	
	public static int getCollsionHandling()
	{
		return collisionHandling;
	}
	
	
	public static long getTime()
	{
		return time;
	}
	
	public static Quat4d makeQFromEulerAngles(double x, double y, double z)
	{
		Quat4d q;
		
		double cx, cy, cz, sx, sy, sz;
		double ccy, ssy, csy, scy;
		
		cz = Math.cos(0.5 * z);
		cy = Math.cos(0.5 * y);
		cx = Math.cos(0.5 * x);
		
		sz = Math.sin(0.5 * z);
		sy = Math.sin(0.5 * y);
		sx = Math.sin(0.5 * x);
		
		ccy = cz * cy;
		ssy = sz * sy;
		csy = cz * sy;
		scy = sz * sy;
		
		q = new Quat4d((ccy * sx - ssy * cx),
						(csy * cx + scy * sx),
						(scy * cx - csy * sx),
						(ccy * cx + ssy * sx));
		
		return q;
	}
	
	public static Vector3d makeEulerAnglesFromQ(Quat4d q)
	{
		double r11,r21,r31,r32,r33,r12,r13;
		double q00,q11,q22,q33;
		double tmp;
		Vector3d u = new Vector3d(0.0,0.0,0.0);
		
		q00 = q.w * q.w;
		q11 = q.x * q.x;
		q22 = q.y * q.y;
		q33 = q.z * q.z;
		
		r11 = q00 + q11 - q22 - q33;
		r21 = 2 * (q.x * q.y + q.w * q.z);
		r31 = 2 * (q.x * q.z - q.w * q.y);
		r32 = 2 * (q.y * q.z + q.w * q.x);
		r33 = q00 - q11 - q22 + q33;
		
		tmp = Math.abs(r31);
		
		if(tmp > 0.999999)
		{
			r12 = 2 * (q.x * q.y - q.w * q.z);
			r13 = 2 * (q.x * q.z + q.w * q.y);
			
			u.x = 0.0;
			u.y = (-(Math.PI / 2) * r31 / tmp);
			u.z = (Math.atan2(-r12, -r31*r13));
			
			return u;
		}
	
		u.x = Math.atan2(r32, r33);
		u.y = Math.asin(-r31);
		u.z = Math.atan2(r21, r11);
		
		return u;
	
	}
	
	//////////////////////////////////////////
	
	private class Point3DAdjacent
	{
		Point3d point = null;
		int key[] = null;
		
		public boolean equals(Object p)
		{
			if(p instanceof Point3d)
			{
				Point3d pt = (Point3d)p;
				return point.x == pt.x && point.y == pt.y && point.z == pt.z;
			}
			return false;
		}
	}
	
	private class Feature
	{
		Point3d vertex = null;
		int[] vertexIndex = null;
		
		Point3d[] face = null;
		Point3d[] line = null;
		
		public Feature(Point3d v, int index)
		{	
			setVertex(v, index);
		}
		
		public Feature(Point3d p1, Point3d p2, Point3d p3, int i1, int i2, int i3)
		{
			setFace(p1,p2,p3, i1, i2, i3);
		}
		public Feature(Point3d p1, Point3d p2, int i1, int i2)
		{
			setLine(p1,p2, i1, i2);
		}
		
		public void setVertex(Point3d v, int index)
		{
			
			vertex = v;
			vertexIndex = new int[1];
			vertexIndex[0] = index;
			face = null;
			line = null;
		}
		
		public void setFace(Point3d p1, Point3d p2, Point3d p3, int i1, int i2, int i3)
		{
			vertex = null;
			vertexIndex = new int[3];
			vertexIndex[0] = i1;
			vertexIndex[1] = i2;
			vertexIndex[2] = i3;
			
			face = new Point3d[3];
			face[0] = p1; face[1] = p2; face[2] = p3;
			line = null;
		}
		
		public void setLine(Point3d p1, Point3d p2, int i1, int i2)
		{
			vertex = null;
			vertexIndex = new int[2];
			vertexIndex[0] = i1;
			vertexIndex[1] = i2;
			
			face = null;			
			line = new Point3d[2];
			line[0] = p1; line[1] = p2;
		}
		
		public Point3d getPoint()
		{
			
			
				
			Point3d result = new Point3d(0.0,0.0,0.0);
			if(vertex != null)
				result = (Point3d)vertex.clone();
			if(face != null)
			{
				result.add(face[0]);
				result.add(face[1]);
				result.add(face[2]);
				result.scale(1.0/face.length);
			}
			else if(line != null)
			{
				result.add(line[0]);
				result.add(line[1]);
				result.scale(1.0/line.length);
			}
			
			Transform3D t = new Transform3D();
			object3D.getTransform(t);
			t.transform(result);
			return result;
		}
		
		public double distanceTo(Feature f)
		{
			return getPoint().distance(f.getPoint());
		}
		
		public double distanceTo(Vector3d v)
		{
			Point3d p = new Point3d(v);
			return getPoint().distance(p);
		}
		
		
		
		public Feature[] getAdjacent()
		{
		
			Feature [] result = null;
			
			if(vertex != null)
			{
				Point3DAdjacent p3a = (Point3DAdjacent)p3dadj.get(vertexIndex[0]);
				result = new Feature[p3a.key.length / 2 + p3a.key.length];
				for(int i = 0; i < p3a.key.length; i += 2)
				{
					result[i / 2] = new Feature(vertex, ((Point3DAdjacent)p3dadj.get(p3a.key[i])).point, 
												((Point3DAdjacent)p3dadj.get(p3a.key[i + 1])).point, vertexIndex[0], p3a.key[i], p3a.key[i+1]);
				}
				for(int i = p3a.key.length / 2; i < p3a.key.length*1.5; i++)
				{
					result[i] = new Feature(vertex, ((Point3DAdjacent)p3dadj.get(p3a.key[i - p3a.key.length / 2])).point, vertexIndex[0], p3a.key[i - p3a.key.length / 2]);
				}
			}
			else if(face != null)
			{
				result = new Feature[3];
				for(int i = 0; i < 3; i++)
					result[i] = new Feature(((Point3DAdjacent)p3dadj.get(vertexIndex[i])).point, vertexIndex[i]);
			}
			else if(line != null)
			{
				result = new Feature[2];
				for(int i = 0; i < 2; i++)
					result[i] = new Feature(((Point3DAdjacent)p3dadj.get(vertexIndex[i])).point, vertexIndex[i]);
			}
			
			return result;
		}
		
		public Vector3d getNormal()
		{
			Vector3d ret = null;
			if(vertex != null)
			{
				 ret = new Vector3d(getPoint());
				 ret.sub(position);
			}
			else if(face != null)
			{
				ret = new Vector3d(face[0]);
				ret.sub(face[1]);
				
				Vector3d v2 = new Vector3d(face[0]);
				v2.sub(face[2]);
				
				ret.cross(ret, v2);
			}
			else
			{
				ret = new Vector3d(getPoint());
				ret.sub(position);
			}
			
			return ret;
		}
		
		
		
/*		public static Vector3d calculateNormal(Feature f1, Feature f2)
		{
			if(f1.face != null)
				return f1.getNormal();
			else if(f2.face != null)
				return f2.getNormal();
				
			
			
		}*/
		
	}
	
}